
import asyncio, curses, sys, pathlib, time

from datetime import timezone

from decimal import Decimal

from curses import ascii

import logging
logger = logging.getLogger(__name__)

from ..propertymembers import getfloat

def shorten(text, width=0, placeholder="..."):
    "Shorten text"
    txt = text.replace("\n", " ")
    if not width:
        return txt
    if len(txt)<=width:
        return txt
    if width <= len(placeholder):
        return placeholder[:width]
    return txt[:width - len(placeholder)] + placeholder


def localtimestring(t):
    "Return a string of the local time (not date)"
    localtime = t.astimezone(tz=None)
    # convert microsecond to integer between 0 and 100
    ms = localtime.microsecond//10000
    return f"{localtime.strftime('%H:%M:%S')}.{ms:0>2d}"


def drawmessage(window, message, bold=False, maxcols=None):
    """Shows message, message is either a text string, or a tuple of (timestamp, message text)"""
    window.clear()
    if not maxcols:
        maxrows, maxcols = window.getmaxyx()
    maxcols = maxcols - 5
    if not isinstance(message, str):
        message = localtimestring(message[0]) + "  " + message[1]

    if len(message) > maxcols:
        messagetoshow = "    " + shorten(message, width=maxcols, placeholder="...")
    else:
        messagetoshow = "    " + message.ljust(maxcols)

    if bold:
        window.addstr(0, 0, messagetoshow, curses.A_BOLD)
    else:
        window.addstr(0, 0, messagetoshow)


def draw_timestamp_state(control, window, vector):
    "Adds the vector timestamp, and its state to the window"
    maxrows, maxcols = window.getmaxyx()
    window.clear()

    state = vector.state
    window.addstr(0, 1, localtimestring(vector.timestamp))

    lowerstate = state.lower()
    if lowerstate == "idle":
        text = "  Idle  "
    elif lowerstate == "ok":
        text = "  OK    "
    elif lowerstate == "busy":
        text = "  Busy  "
    elif lowerstate == "alert":
        text = "  Alert "
    else:
        return
    window.addstr(0, maxcols - 20, text, control.color(state))


class Button:

    def __init__(self, window, btntext, row, col, btnlen=None, onclick=None):
        self.window = window
        self.btntext = btntext
        self.row = row
        self.col = col
        self.onclick = onclick
        self._focus = False
        self._show = True
        self.bold = False
        if btnlen:
            self.btnlen = btnlen
        else:
            # no btnlen given
            self.btnlen = len(self.btntext) + 2
        # Set this True to make the whole button reverse when in focus
        self.focusreverse = True
        # if it is False, then only the [] will go bold on focus


    def __contains__(self, mouse):
        "Returns True if the mouse y, x are within this field"
        if not self._show:
            return False
        originrow, origincol = self.window.getbegyx()
        if mouse[0] != originrow + self.row:
            return False
        if mouse[1] < origincol + self.col:
            return False
        if mouse[1] < origincol + self.col + self.btnlen:
            return True
        return False

    @property
    def show(self):
        return self._show

    @show.setter
    def show(self, value):
        # setting show False, also sets focus False
        if not value:
            self._focus = False
        self._show = value

    @property
    def focus(self):
        return self._focus

    @focus.setter
    def focus(self, value):
        if not self._show:
            # focus can only be set if show is True
            return
        self._focus = value

    def text(self):
        "pad out the text to be drawn"
        if len(self.btntext) == self.btnlen-2:
            return self.btntext
        if len(self.btntext) > self.btnlen-2:
            return shorten(self.btntext, width=self.btnlen-2, placeholder="...")
        spaces = self.btnlen-2-len(self.btntext)
        front = spaces//2
        back = spaces-front
        return " "*front + self.btntext + " "*back


    def draw(self):
        if not self._show:
            self.window.addstr( self.row, self.col, " "*self.btnlen)
            return
        if self._focus:
            if self.focusreverse:
                self.window.addstr( self.row, self.col, f"[{self.text()}]", curses.A_REVERSE)
            elif self.bold:
                # both in focus and bold
                self.window.addstr( self.row, self.col, f"[{self.text()}]", curses.A_BOLD)
            else:
                self.window.addstr( self.row, self.col, "[", curses.A_BOLD)
                self.window.addstr( self.row, self.col+1, f"{self.text()}")
                self.window.addstr( self.row, self.col+self.btnlen-1, "]", curses.A_BOLD)
        elif self.bold:
            # bold, but not in focus
            self.window.addstr( self.row, self.col, "[")
            self.window.addstr( self.row, self.col+1, f"{self.text()}", curses.A_BOLD)
            self.window.addstr( self.row, self.col+self.btnlen-1, "]")
        else:
            self.window.addstr( self.row, self.col, f"[{self.text()}]")

    def alert(self):
        "draw the button with a red background and INVALID Message"
        if not self._show:
            return
        self.window.addstr( self.row, self.col, f"[{self.text()}] INVALID!", curses.color_pair(3))

    def ok(self):
        "draw the button with a green background and OK Message"
        if not self._show:
            return
        self.window.addstr( self.row, self.col, f"[{self.text()}] OK!", curses.color_pair(1))


class Text:

    """An object which creates and draws a text input field.
       Has a contains method so can be used to check if a mouse
       is in the field and a focus property to indicate the field
       is in focus.
       has methods to accept a key, and return the text"""

    def __init__(self, stdscr, window, text, row, col, txtlen=None):
        self.stdscr = stdscr
        self.window = window
        self.row = row
        self.col = col
        self._focus = False
        self._show = True
        if txtlen:
            # txtlen includes the two [ ] brackets
            self.txtlen = txtlen
        else:
            # no txtlen given
            self.txtlen = len(text) + 2

        self.text = text

        # create a new texteditor
        self.new_texteditor()


    @property
    def text(self):
        return self._text

    @text.setter
    def text(self, text):
        # set text into self._text
        if len(text) > self.txtlen-2:
            self._text = shorten(text, width=self.txtlen-2, placeholder="...")
        elif len(text) == self.txtlen-2:
            self._text = text
        else:
            self._text = text.ljust(self.txtlen-2)  # pads space to right of the text


    def __contains__(self, mouse):
        "Returns True if the mouse y, x are within this field"
        if not self._show:
            return False
        originrow, origincol = self.window.getbegyx()
        if mouse[0] != originrow + self.row:
            return False
        if mouse[1] < origincol + self.col:
            return False
        if mouse[1] < origincol + self.col + self.txtlen:
            return True
        return False


    @property
    def show(self):
        return self._show

    @show.setter
    def show(self, value):
        # setting show False, also sets focus False
        if not value:
            self.focus = False
        self._show = value

    @property
    def focus(self):
        return self._focus

    @focus.setter
    def focus(self, value):
        if not self._show:
            # focus can only be set if show is True
            return
        if value:
            # setting focus on, shows the cursor
            curses.curs_set(1)
            self._focus = True
        else:
            # setting focus off, hides the cursor
            curses.curs_set(0)
            self._focus = False

    def draw(self):
        if not self._show:
            self.window.addstr( self.row, self.col, " "*self.txtlen)
            return
        if self._focus:
            self.window.addstr( self.row, self.col, "[", curses.A_BOLD)
            self.window.addstr( self.row, self.col+1, self._text)
            self.window.addstr( self.row, self.col+self.txtlen-1, "]", curses.A_BOLD)
        else:
            self.window.addstr( self.row, self.col, "["+self._text)
            self.window.addstr( self.row, self.col+self.txtlen-1, "]")

    def movecurs(self):
        "Moves the cursor including spaces"
        self._editstring.movecurs()

    def new_texteditor(self):
        originrow, origincol = self.window.getbegyx()
        self._editstring = EditString(self.stdscr,
                            originrow + self.row,                     # row
                            origincol + self.col + 1,                 # start col
                            origincol + self.col + self.txtlen - 2,   # endcol
                            self.text )                               # the actual text

    def gettext(self, key):
        "called with each keypress, returns new text"
        newvalue = self._editstring.gettext(key)
        self.text = newvalue
        return newvalue


    def getnumber(self, key):
        "called with each keypress, returns new number string"
        newvalue = self._editstring.getnumber(key)
        self.text = newvalue
        return newvalue


class EditString():

    """Object to edit characters in a field, has gettext(key) and
       getnumber(key) methods which given a key press, edits and returns
       the fields string"""

    def __init__(self, stdscr, row, startcol, endcol, text):
        "Class to input text"
        self.stdscr = stdscr
        self.row = row
        self.startcol = startcol
        self.endcol = endcol
        self.length = endcol - startcol + 1
        if len(text) > self.length:
            self.text = text[:self.length]
        else:
            # pad text with right hand spaces
            self.text = text.ljust(self.length)
        # put curser at end of text
        self.stringpos = len(self.text.rstrip())
        self.movecurs()

    def insertch(self, ch):
        "Insert a character at stringpos"
        if self.stringpos >= self.length:
            # stringpos must be less than the length
            return
        self.text = self.text[:self.stringpos] + ch + self.text[self.stringpos:-1]
        self.stringpos += 1

    def delch(self):
        "delete character at stringpos-1"
        if not self.stringpos:
            # stringpos must be greater than zero
            return
        self.text = self.text[:self.stringpos-1] + self.text[self.stringpos:] + " "
        self.stringpos -= 1

    def movecurs(self):
        self.stdscr.move(self.row, self.startcol+self.stringpos)
        self.stdscr.refresh()

    def gettext(self, key):
        "called with each keypress, returns new text"
        if ascii.isprint(key):
            if self.stringpos >= self.length:
                # at max length, return
                return self.text
            ch = chr(key)
            self.insertch(ch)
        elif key>255:
            # control character
            if ((key == curses.KEY_DC) or (key == curses.KEY_BACKSPACE)) and self.stringpos:
                # delete character (self.stringpos cannot be zero)
                self.delch()
            elif (key == curses.KEY_LEFT) and self.stringpos:
                # move cursor left (self.stringpos cannot be zero)
                self.stringpos -= 1
            elif (key == curses.KEY_RIGHT) and (self.stringpos < self.length):
                # move cursor right
                self.stringpos += 1
        return self.text

    def getnumber(self, key):
        "called with each keypress, returns new number string"
        if ascii.isdigit(key):
            if self.stringpos >= self.length:
                # at max length, return
                return self.text
            ch = chr(key)
            self.insertch(ch)
        elif ascii.isprint(key):
            if self.stringpos >= self.length:
                # at max length, return
                return self.text
            ch = chr(key)
            if ch in (".", " ", ":", ";", "-", "+"):
                self.insertch(ch)
        elif key>255:
            # control character
            if ((key == curses.KEY_DC) or (key == curses.KEY_BACKSPACE)) and self.stringpos:
                # delete character (self.stringpos cannot be zero)
                self.delch()
            elif (key == curses.KEY_LEFT) and self.stringpos:
                # move cursor left (self.stringpos cannot be zero)
                self.stringpos -= 1
            elif (key == curses.KEY_RIGHT) and (self.stringpos < self.length):
                # move cursor right
                self.stringpos += 1
        return self.text


class BaseMember:

    def __init__(self, stdscr, control, window, tstatewin, vector, name, namelen=0):

        self.stdscr = stdscr
        self.control = control
        self.client = control.client
        self.window = window
        self.tstatewin = tstatewin
        self.vector = vector
        self.name = name
        self.maxrows, self.maxcols = self.window.getmaxyx()
        # namelen sets a button on the left with length
        # if namelen not given, or too long, sets max len of self.maxcols//2-10
        # that is, for a widow of 80, this gives button text of 30, button size = 32 (for two brackets)
        if (not namelen) or (namelen>self.maxcols//2-10):
            namelen = self.maxcols//2-10
        membersdict = self.vector.members()
        self.member = membersdict[name]
        # self.member is a propertymember
        self.startline = 0
        # linecount is the number of lines this widget takes up,
        # including an end empty line
        self.linecount = 4
        self.name_btn = Button(window, self.name, 0, 1, namelen+2)
        self._focus = False
        # if close string is set, it becomes the return value from input routines
        self._close = ""
        # self._newvalue is the value edited
        self._newvalue = ""

    def close(self, value):
        self._close = value

    def value(self):
        return self.vector[self.name]

    def reset(self):
        "Reset the widget removing any value updates, called by cancel"
        pass

    @property
    def focus(self):
        return self._focus

    @focus.setter
    def focus(self, value):
        if self._focus == value:
            return
        self._focus = value
        self.name_btn.focus = value


    @property
    def endline(self):
        "self.endline is the empty line after the vector"
        return self.startline + self.linecount - 1

    def draw(self, startline=None):
        if not startline is None:
            self.startline = startline
        displaylabel = shorten(self.vector.memberlabel(self.name), width=self.maxcols-5, placeholder="...")
        self.window.addstr( self.startline, 1, displaylabel, curses.A_BOLD )
        self.name_btn.row = self.startline+1
        self.name_btn.draw()

    def newvalue(self):
        return self._newvalue


    async def setkey(self, key):
        return key

    async def keyinput(self):
        """Waits for a key press,
           if self.control.stop is True, returns 'Stop',
           if screen has been resized, returns 'Resize',
           if self._close has been given a value, returns that value
           Otherwise returns the key pressed, or a mouse tuple
           if the mouse has been pressed."""
        while True:
            if self.control.stop:
                return "Stop"
            if self._close:
                return self._close
            key = self.stdscr.getch()
            if key == -1:
                await asyncio.sleep(0.02)
                continue
            if key == curses.KEY_RESIZE:
                return "Resize"
            if key == curses.KEY_MOUSE:
                try:
                    mouse = curses.getmouse()
                except curses.error:
                    continue
                # mouse is (id, x, y, z, bstate)
                if mouse[4] == curses.BUTTON1_RELEASED:
                    # return a tuple of the mouse coordinates
                    #          row     col
                    return (mouse[2], mouse[1])
                await asyncio.sleep(0)
                continue
            return key




class SwitchMember(BaseMember):

    def __init__(self, stdscr, control, window, tstatewin, vector, name, namelen=0):
        super().__init__(stdscr, control, window, tstatewin, vector, name, namelen)
        # create  ON, OFF buttons
        self.on = Button(window, 'ON', 0, 0)
        self.on.focusreverse = False
        self.on.bold = True if self.value() == "On" else False
        self.off = Button(window, 'OFF', 0, 0)
        self.off.focusreverse = False
        self.off.bold = not self.on.bold
        self.linecount = 3

    def reset(self):
        "Reset the widget removing any value updates, called by cancel"
        if self.vector.perm == "ro":
            return
        # Draw the on/off buttons
        self.on.bold = True if self.value() == "On" else False
        self.on.row = self.startline+1
        self.on.col = self.maxcols-15
        self.on.draw()
        self.off.bold = not self.on.bold
        self.off.row = self.startline+1
        self.off.col = self.maxcols-10
        self.off.draw()

    def draw(self, startline=None):
        super().draw(startline)
        # draw the On or Off value
        self.window.addstr( self.startline+1, self.maxcols-20, self.value(), curses.A_BOLD )
        if self.vector.perm == "ro":
            return
        # Draw the on/off buttons
        self.on.row = self.startline+1
        self.on.col = self.maxcols-15
        self.on.draw()
        self.off.bold = not self.on.bold
        self.off.row = self.startline+1
        self.off.col = self.maxcols-10
        self.off.draw()


    def newvalue(self):
        if self.on.bold:
            return "On"
        return "Off"

    @property
    def focus(self):
        return self._focus

    @focus.setter
    def focus(self, value):
        if self._focus == value:
            return
        self._focus = value
        self.name_btn.focus = value
        self.on.focus = False
        self.off.focus = False

    def set_on_focus(self):
        self._focus = True
        self.name_btn.focus = False
        self.name_btn.draw()
        self.on.focus = True
        self.on.draw()
        self.off.focus = False
        self.off.draw()

    def set_off_focus(self):
        self._focus = True
        self.name_btn.focus = False
        self.name_btn.draw()
        self.on.focus = False
        self.on.draw()
        self.off.focus = True
        self.off.draw()


    async def handlemouse(self, key):
        """Handles a mouse input
           Sets On, Off buttons bold as appropriate.
           Returns None if mouse not on any button
                   or if on name, and already focused, and ro
           Returns 'focused' if on any button to indicate
                   a button has focus
           Returns "set_on" if on chosen and rule requires this to be only button"""
        if key in self.name_btn:
            if self.name_btn.focus:
                return "Member"
            else:
                self._focus = True
                self.name_btn.focus = True
                self.name_btn.draw()
                self.on.focus = False
                self.on.draw()
                self.off.focus = False
                self.off.draw()
                return "focused"
        if key in self.on:
            if self.on.focus:
                self.on.bold = True
                self.off.bold = False
                self.on.draw()
                self.off.draw()
                if self.vector.rule != 'AnyOfMany':  ### 'OneOfMany','AtMostOne' both require this to be the only on button
                    return "set_on"
                return "focused"
            else:
                self._focus = True
                self.on.focus = True
                self.on.draw()
                self.off.focus = False
                self.off.draw()
                self.name_btn.focus = False
                self.name_btn.draw()
                return "focused"
        if key in self.off:
            if self.off.focus:
                self.off.bold = True
                self.on.bold = False
                self.on.draw()
                self.off.draw()
                return "focused"
            else:
                self._focus = True
                self.off.focus = True
                self.off.draw()
                self.on.focus = False
                self.on.draw()
                self.name_btn.focus = False
                self.name_btn.draw()
                return "focused"


# 32 space, 9 tab, 353 shift tab, 261 right arrow, 260 left arrow, 10 return, 339 page up, 338 page down, 259 up arrow, 258 down arrow

    async def setkey(self, key):
        """This deals with keystroke inputs - not mouse input
           If no button is in focus, returns None
           if next/prev widget, returns key
           Returns None if key not relevant or if
           a focus change can be handled here, and no further
           action is necessary.
           Returns "set_on" if on chosen and rule requires this to be only button
           """

        if self.name_btn.focus:
            if key in (353, 260, 339, 338, 259, 258):  # 353 shift tab, 260 left arrow, 339 page up, 338 page down, 259 up arrow, 258 down arrow
                # go to next or previous member widget
                return key
            if key == 10:     # 10 return
                return "Member"
            if key in (32, 9, 261):     # 32 space, 9 tab, 261 right arrow,
                # go to on button
                self.name_btn.focus = False
                self.on.focus = True
                self.name_btn.draw()
                self.on.draw()
                self.window.noutrefresh()
                curses.doupdate()
            # ignore any other key
            return
        elif self.on.focus:
            if key == 10:                  # 10 return
                # set on key as bold, off key as standard
                self.on.bold = True
                self.off.bold = False
                self.on.draw()
                self.off.draw()
                if self.vector.rule != 'AnyOfMany':  ### 'OneOfMany','AtMostOne' both require this to be the only on button
                    return "set_on"
            elif key in (339, 259):   # 339 page up, 259 up arrow
                return "onup"
            elif key in (338, 258):   # 338 page down, 258 down arrow
                return "ondown"
            elif key in (353, 260): # 353 shift tab, 260 left arrow
                # back to name_btn
                self.name_btn.focus = True
                self.on.focus = False
                self.name_btn.draw()
                self.on.draw()
            elif key in (32, 9, 261):  # 32 space, 9 tab, 261 right arrow
                # move to off btn
                self.on.focus = False
                self.on.draw()
                self.off.focus = True
                self.off.draw()
            else:
                return
            self.window.noutrefresh()
            curses.doupdate()
            return
        elif self.off.focus:
            if key == 10:                      # 10 return
                # set off key as bold, on key as standard
                self.off.bold = True
                self.on.bold = False
                self.off.draw()
                self.on.draw()
            elif key in (339, 259):   # 339 page up, 259 up arrow
                return "offup"
            elif key in (338, 258):   # 338 page down, 258 down arrow
                return "offdown"
            elif key in (32, 261):   # 32 space, 261 right arrow
                # go to name_btn
                self.off.focus = False
                self.off.draw()
                self.name_btn.focus = True
                self.name_btn.draw()
            elif key == 9:   # 9 tab
                # go to next widget or scroll pad
                self.off.focus = False
                self.off.draw()
                return key
            elif key in (353, 260):  # 353 shift tab, 260 left arrow
                # back to on btn
                self.off.focus = False
                self.off.draw()
                self.on.focus = True
                self.on.draw()
            else:
                return
            self.window.noutrefresh()
            curses.doupdate()



class LightMember(BaseMember):

    def __init__(self, stdscr, control, window, tstatewin, vector, name, namelen=0):
        super().__init__(stdscr, control, window, tstatewin, vector, name, namelen)
        self.linecount = 3


    def draw(self, startline=None):
        super().draw(startline)
        # draw the light value
        lowervalue = self.value().lower()
        if lowervalue == "idle":
            text = "  Idle  "
        elif lowervalue == "ok":
            text = "  OK    "
        elif lowervalue == "busy":
            text = "  Busy  "
        elif lowervalue == "alert":
            text = "  Alert "
        else:
            return
        # draw the value
        self.window.addstr(self.startline+1, self.maxcols-20, text, self.control.color(lowervalue))

    async def setkey(self, key):
        "This widget is in focus, but is read only"
        return key

    async def handlemouse(self, key):
        "Handles a mouse input"
        if key in self.name_btn:
            if self.name_btn.focus:
                return "Member"
            else:
                self._focus = True
                self.name_btn.focus = True
                self.name_btn.draw()
                return "focused"



#   <!ATTLIST defNumberVector
#   device %nameValue; #REQUIRED        name of Device
#   name %nameValue; #REQUIRED          name of Property
#   label %labelValue; #IMPLIED         GUI label, use name by default
#   group %groupTag; #IMPLIED           Property group membership, blank by default
#   state %propertyState; #REQUIRED     current state of Property
#   perm %propertyPerm; #REQUIRED       ostensible Client controlability
#   timeout %numberValue; #IMPLIED      worse-case time to affect, 0 default, N/A for ro
#   timestamp %timeValue #IMPLIED       moment when these data were valid
#   message %textValue #IMPLIED         commentary

#   Define one member of a number vector
#   <!ATTLIST defNumber
#   name %nameValue; #REQUIRED          name of this number element
#   label %labelValue; #IMPLIED         GUI label, or use name by default
#   format %numberFormat; #REQUIRED     printf-style format for GUI display
#   min %numberValue; #REQUIRED         minimal value
#   max %numberValue; #REQUIRED         maximum value, ignore if min == max
#   step %numberValue; #REQUIRED        allowed increments, ignore if 0


class NumberMember(BaseMember):

    def __init__(self, stdscr, control, window, tstatewin, vector, name, namelen=0):
        super().__init__(stdscr, control, window, tstatewin, vector, name, namelen)
        if self.vector.perm == "ro":
            self.linecount = 3
        else:
            self.linecount = 4
        # the newvalue to be edited and sent
        self._newvalue = self.member.getformattedvalue()
        # create a string input field
                                     # window         text        row               col,               length of field
        self.edit_txt = Text(stdscr, self.window, self._newvalue, self.startline+2, self.maxcols//2+2, txtlen=self.maxcols//2 - 6)


    @property
    def focus(self):
        return self._focus

    @focus.setter
    def focus(self, value):
        "Sets focus to be either False, or if True, set the name_btn focus as True"
        self._focus = value
        self.name_btn.focus = value
        # and regardless of value, set self.edit_txt.focus to False, but check number ok
        if self.edit_txt.focus:
            self.checknumber()
            self.edit_txt.text = self._newvalue
            self.edit_txt.focus = False

    def set_edit_focus(self):
        self._focus = True
        self.name_btn.focus = False
        self.name_btn.draw()
        self.edit_txt.focus = True
        self.edit_txt.draw()


    def reset(self):
        "Reset the widget removing any value updates, called by cancel"
        if self.vector.perm == "ro":
            return
        self._newvalue = self.member.getformattedvalue()
        # draw the value to be edited
        self.edit_txt.text = self._newvalue
        self.edit_txt.draw()


    def draw(self, startline=None):
        super().draw(startline)
        txtlen=self.maxcols//2 - 6

        # draw the number value
        text = self.member.getformattedvalue().strip()
        if len(text) > txtlen:
            text = text[:txtlen]
        else:
            text = text.ljust(txtlen)
        # draw the value
        self.window.addstr(self.startline+1, self.maxcols//2+3, text, curses.A_BOLD)

        if self.vector.perm == "ro":
            return
        self.edit_txt.row = self.startline+2
        self.edit_txt.draw()


    async def handlemouse(self, key):
        """Handles a mouse input
           Returns None if mouse not in name btn or edit field
                   or if in name, but perm is ro
                   or if mouse in edit field which already has focus

           If mouse in name button and not focused return focused
           If mouse in name button and focused, make edit field focus and return edit
           If mouse in edit field without focus, set focus and return 'edit'
            """
        if key in self.name_btn:
            if self.name_btn.focus:
                return "Member"
            else:
                self._focus = True
                self.name_btn.focus = True
                self.name_btn.draw()
                if self.edit_txt.focus:
                    self.checknumber()
                    self.edit_txt.text = self._newvalue
                    self.edit_txt.focus = False
                    self.edit_txt.draw()
                return "focused"
        if key in self.edit_txt:
            if self.edit_txt.focus:
                # already in focus, do nothing
                return
            else:
                self.set_edit_focus()
                return "edit"


    async def setkey(self, key):
        """Handles key input if name button is in focus
           Return None if name not in focus
           Return key for next/prev widget
           Return 'edit' if moving to edit field"""
        if self.name_btn.focus:
            if key in (353, 260, 339, 338, 259, 258):  # 353 shift tab, 260 left arrow, 339 page up, 338 page down, 259 up arrow, 258 down arrow
                # go to next or previous member widget
                return key
            if key == 10:     # 10 return
                return "Member"
            if key in (9, 32, 261):     # 9 tab, 32 space, 261 right arrow
                self.set_edit_focus()
                self.window.noutrefresh()
                curses.doupdate()
                return "edit"
        else:
            # name button not in focus, so this key must come from the editable field
            if key in (339, 259):  # 339 page up, 259 up arrow
                # go previous member widget edit field
                return "editup"
            if key in (338, 258, 10):  # 338 page down, 258 down arrow, 10 enter
                # go to next widget edit field
                return "editdown"
            if key == 9:       # 9 tab
                return 9



    async def inputfield(self):
        """Input number, set it into self._newvalue
           If mouse pressed outside edit field, return the mouse tuple"""
        self.edit_txt.new_texteditor()

        while not self.control.stop:
            key = await self.keyinput()
            if key in ("Resize", "Messages", "Devices", "Vectors", "Stop"):
                return key
            if isinstance(key, tuple):
                if key in self.edit_txt:
                    continue
                return key
            if key == 353:       # 353 shift tab
                self.checknumber()
                self.name_btn.focus = True
                self.name_btn.draw()
                self.edit_txt.text = self._newvalue
                self.edit_txt.focus = False
                self.edit_txt.draw()
                self.window.noutrefresh()
                curses.doupdate()
                return
            if key in (10, 9, 339, 338, 259, 258):    # 10 enter, 9 tab, 339 page up, 338 page down, 259 up arrow, 258 down arrow
                self.checknumber()
                self.edit_txt.text = self._newvalue
                return key
            self._newvalue = self.edit_txt.getnumber(key)
            self.edit_txt.draw()
            self.window.noutrefresh()
            self.edit_txt.movecurs()
            curses.doupdate()


    def checknumber(self):
        "set self._newvalue, limiting it to correct range"
        # self._newvalue is the new value input
        try:
            newfloat = getfloat(self._newvalue)
        except (ValueError, TypeError):
            # reset self._newvalue
            self._newvalue = self.member.getformattedvalue()
            return
        # check step, and round newfloat to nearest step value
        stepvalue = getfloat(self.member.step)
        minvalue = getfloat(self.member.min)
        if stepvalue:
            stepvalue = Decimal(str(stepvalue))
            difference = newfloat - minvalue
            newfloat = minvalue + float(int(Decimal(str(difference)) / stepvalue) * stepvalue)
        # check not less than minimum
        if newfloat < minvalue:
            # reset self._newvalue to be the minimum, and accept this
            self._newvalue = self.member.getformattedstring(minvalue)
            return
        if self.member.max != self.member.min:
            maxvalue = getfloat(self.member.max)
            if newfloat > maxvalue:
                # reset self._newvalue to be the maximum, and accept this
                self._newvalue = self.member.getformattedstring(maxvalue)
                return
        # reset self._newvalue to the correct format, and accept this
        self._newvalue = self.member.getformattedstring(newfloat)


# <!ATTLIST defTextVector
# device %nameValue; #REQUIRED   name of Device
# name %nameValue; #REQUIRED     name of Property
# label %labelValue; #IMPLIED    GUI label, use name by default
# group %groupTag; #IMPLIED      Property group membership, blank by default
# state %propertyState; #REQUIRED  current state of Property
# perm %propertyPerm; #REQUIRED    ostensible Client controlability
# timeout %numberValue; #IMPLIED   worse-case time to affect, 0 default, N/A for ro
# timestamp %timeValue #IMPLIED    moment when these data were valid
# message %textValue #IMPLIED      commentary

# Define one member of a text vector
# <!ELEMENT defText %textValue >
# <!ATTLIST defText
# name %nameValue; #REQUIRED
# label %labelValue; #IMPLIED

class TextMember(BaseMember):

    def __init__(self, stdscr, control, window, tstatewin, vector, name, namelen=0):
        super().__init__(stdscr, control, window, tstatewin, vector, name, namelen)
        if self.vector.perm == "ro":
            self.linecount = 3
        else:
            self.linecount = 4
        # the newvalue to be edited and sent
        self._newvalue = self.vector[self.name]

                                     # window         text        row                col           length of field
        self.edit_txt = Text(stdscr, self.window, self._newvalue, self.startline+2, self.maxcols//2+2, txtlen=self.maxcols//2 - 6)

    @property
    def focus(self):
        return self._focus

    @focus.setter
    def focus(self, value):
        if self._focus == value:
            return
        self._focus = value
        self.name_btn.focus = value
        self.edit_txt.focus = False

    def set_edit_focus(self):
        self._focus = True
        self.name_btn.focus = False
        self.name_btn.draw()
        self.edit_txt.focus = True
        self.edit_txt.draw()


    def reset(self):
        "Reset the widget removing any value updates, called by cancel"
        if self.vector.perm == "ro":
            return
        self._newvalue = self.member.membervalue
        # draw the value to be edited
        self.edit_txt.text = self._newvalue
        self.edit_txt.draw()

    def draw(self, startline=None):
        super().draw(startline)

        txtlen=self.maxcols//2 - 6

        # draw the text
        text = self.member.membervalue
        if len(text) > txtlen:
            text = text[:txtlen]
        else:
            text = text.ljust(txtlen)
        # draw the value
        self.window.addstr(self.startline+1, self.maxcols//2+3, text, curses.A_BOLD)

        if self.vector.perm == "ro":
            return
        self.edit_txt.row = self.startline+2
        self.edit_txt.draw()


    async def handlemouse(self, key):
        "Handles a mouse input"
        if key in self.name_btn:
            if self.name_btn.focus:
                return "Member"
            else:
                self._focus = True
                self.name_btn.focus = True
                self.name_btn.draw()
                if self.edit_txt.focus:
                    self.edit_txt.text = self._newvalue
                    self.edit_txt.focus = False
                    self.edit_txt.draw()
                return "focused"
        if key in self.edit_txt:
            if self.edit_txt.focus:
                # already in focus, do nothing
                return
            else:
                self.set_edit_focus()
                return "edit"


    async def setkey(self, key):
        "This widget is in focus, and deals with inputs"
        if self.name_btn.focus:
            if key in (353, 260, 339, 338, 259, 258):  # 353 shift tab, 260 left arrow, 339 page up, 338 page down, 259 up arrow, 258 down arrow
                # go to next or previous member widget
                return key
            if key == 10:     # 10 return
                return "Member"
            if key in (9, 32, 261):     # 9 tab, 32 space, 261 right arrow
                self.set_edit_focus()
                self.window.noutrefresh()
                curses.doupdate()
                return "edit"
        else:
            # name button not in focus, so this key must come from the editable field
            if key in (339, 259):  # 339 page up, 259 up arrow
                # go previous member widget edit field
                return "editup"
            if key in (338, 258, 10):  # 338 page down, 258 down arrow, 10 enter
                # go to next widget edit field
                return "editdown"
            if key == 9:
                return 9 # tab key for next item


    async def inputfield(self):
        """Input text, set it into self._newvalue
           If mouse pressed outside edit field, return the mouse tuple"""
        self.edit_txt.new_texteditor()

        while not self.control.stop:
            key = await self.keyinput()
            if key in ("Resize", "Messages", "Devices", "Vectors", "Stop"):
                return key
            if isinstance(key, tuple):
                if key in self.edit_txt:
                    continue
                return key
            if key == 353:       # 353 shift tab
                self.name_btn.focus = True
                self.name_btn.draw()
                self.edit_txt.text = self._newvalue
                self.edit_txt.focus = False
                self.edit_txt.draw()
                self.window.noutrefresh()
                curses.doupdate()
                return
            if key in (10, 9, 339, 338, 259, 258):    # 339 page up, 338 page down, 259 up arrow, 258 down arrow
                self.edit_txt.text = self._newvalue
                return key
            self._newvalue = self.edit_txt.gettext(key)
            self.edit_txt.draw()
            self.window.noutrefresh()
            self.edit_txt.movecurs()
            curses.doupdate()



# Define a property that holds one or more Binary Large Objects, BLOBs.
# <!ELEMENT defBLOBVector (defBLOB+) >
# <!ATTLIST defBLOBVector
# device %nameValue; #REQUIRED       name of Device
# name %nameValue; #REQUIRED         name of Property
# label %labelValue; #IMPLIED        GUI label, use name by default
# group %groupTag; #IMPLIED          Property group membership, blank by default
# state %propertyState; #REQUIRED    current state of Property
# perm %propertyPerm; #REQUIRED      ostensible Client controlability
# timeout %numberValue; #IMPLIED     worse-case time to affect, 0 default, N/A for ro
# timestamp %timeValue #IMPLIED      moment when these data were valid
# message %textValue #IMPLIED        commentary

# Define one member of a BLOB vector. Unlike other defXXX elements, this does not contain an
# initial value for the BLOB.
# <!ELEMENT defBLOB EMPTY >
# <!ATTLIST defBLOB
# name %nameValue; #REQUIRED          name of this BLOB element
# label %labelValue; #IMPLIED         GUI label, or use name by default



class BLOBMember(BaseMember):

    def __init__(self, stdscr, control, window, tstatewin, vector, name, namelen=0):
        super().__init__(stdscr, control, window, tstatewin, vector, name, namelen)
        if self.vector.perm == "ro":
            self.linecount = 3
        else:
            self.linecount = 4
                                     # window         text        row               col,               length of field
        self.edit_txt = Text(stdscr, self.window, self._newvalue, self.startline+2, self.maxcols//2-15, txtlen=self.maxcols//2+4)
        self.send_btn = Button(window, "Send", 0, self.maxcols-9, 6)


    @property
    def focus(self):
        return self._focus

    @focus.setter
    def focus(self, value):
        if self._focus == value:
            return
        self._focus = value
        self.name_btn.focus = value
        self.edit_txt.focus = False
        self.send_btn.focus = False

    def set_edit_focus(self):
        self._focus = True
        self.name_btn.focus = False
        self.name_btn.draw()
        self.send_btn.focus = False
        self.send_btn.draw()
        self.edit_txt.focus = True
        self.edit_txt.draw()

    def set_send_focus(self):
        self._focus = True
        self.name_btn.focus = False
        self.name_btn.draw()
        self.send_btn.focus = True
        self.send_btn.draw()
        self.edit_txt.focus = False
        self.edit_txt.draw()


    def filename(self):
        "Returns filename of last file received and saved"
        nametuple = (self.vector.devicename, self.vector.name, self.name)
        if nametuple in self.control.BLOBfiles:
            return self.control.BLOBfiles[nametuple].name
        else:
            return ""

    def reset(self):
        "Reset the widget removing any value updates, called by cancel"
        if self.vector.perm == "ro":
            return
        self._newvalue = self.member.membervalue
        # draw the value to be edited
        self.edit_txt.text = self._newvalue
        self.edit_txt.draw()


    def draw(self, startline=None):
        super().draw(startline)

        # draw the received file
        if self.vector.perm != "wo":
            # Set the length of the received filename to fit on the screen
            rxfilenamelength = self.maxcols//2 - 2
            rxfile = self.filename()
            if not rxfile:
                rxfile = " "*rxfilenamelength
            else:
                rxfile = shorten(rxfile, width=rxfilenamelength, placeholder="...")
            # draw the value
            self.window.addstr(self.startline+1, self.maxcols//2, rxfile, curses.A_BOLD)

        if self.vector.perm == "ro":
            return

        # Draw the editable field
        self.window.addstr( self.startline+2, 1, "Filepath to send:" )  # 18 characters
        self.edit_txt.row = self.startline+2
        self.edit_txt.draw()

        # send button
        self.send_btn.row = self.startline+2
        self.send_btn.draw()


    async def handlemouse(self, key):
        "Handles a mouse input"
        if key in self.name_btn:
            if self.name_btn.focus:
                return "Member"
            else:
                self._focus = True
                self.name_btn.focus = True
                self.name_btn.draw()
                if self.edit_txt.focus:
                    self.edit_txt.text = self._newvalue
                    self.edit_txt.focus = False
                    self.edit_txt.draw()
                elif self.send_btn.focus:
                    self.send_btn.focus = False
                    self.send_btn.draw()
                return "focused"
        if key in self.edit_txt:
            if self.edit_txt.focus:
                # already in focus, do nothing
                return
            else:
                self.set_edit_focus()
                return "edit"
        if key in self.send_btn:
            if self.send_btn.focus:
                # send the vector with this member, and set focus to name_btn
                self.send_btn.focus = False
                self.send_btn.draw()
                try:
                    filepath = pathlib.Path(self._newvalue.strip()).expanduser().resolve()
                    blobformat = ''.join(filepath.suffixes)
                    members = {self.name : (filepath, 0, blobformat)}
                    await self.vector.send_newBLOBVector(members=members)
                except Exception:
                    logger.exception("Exception report from BLOB send action")
                    self.window.addstr( self.startline+2, 1, "!! Invalid !!    ", curses.color_pair(3) )
                else:
                    self.window.addstr( self.startline+2, 1, " - Sending -     ", curses.color_pair(1) )
                    self.vector.state = 'Busy'
                    draw_timestamp_state(self.control, self.tstatewin, self.vector)
                    self.tstatewin.noutrefresh()
                self.window.noutrefresh()
                curses.doupdate()
                time.sleep(0.4)      # blocking, to avoid screen being changed while this time elapses
                self.window.addstr( self.startline+2, 1, "Filepath to send:" )
                self.window.noutrefresh()
                return "senddown"
            else:
                self._focus = True
                self.name_btn.focus = False
                self.name_btn.draw()
                self.send_btn.focus = True
                self.send_btn.draw()
                self.edit_txt.text = self._newvalue
                self.edit_txt.focus = False
                self.edit_txt.draw()
                return "focused"


    async def setkey(self, key):
        "This widget is in focus, and deals with inputs"
        if self.name_btn.focus:
            if key in (353, 260, 339, 338, 259, 258):  # 353 shift tab, 260 left arrow, 339 page up, 338 page down, 259 up arrow, 258 down arrow
                # go to next or previous member widget
                return key
            if key == 10:     # 10 return
                return "Member"
            if key in (9, 32, 261):     # 9 tab, 32 space, 261 right arrow
                # name_btn is in focus, set edit_txt in focus and return "edit"
                # which informs the window to await inputfield
                self.set_edit_focus()
                self.window.noutrefresh()
                curses.doupdate()
                return "edit"
        elif self.send_btn.focus:
            if key in (9, 261):  # 9 tab, 261 right arrow
                # go to next or previous member widget
                self.send_btn.focus = False
                self.send_btn.draw()
                return key
            if key in (339, 259):  # 339 page up, 259 up arrow
                # go previous member widget send button
                return "sendup"
            if key in (338, 258):  # 338 page down, 258 down arrow
                # go to next widget send button
                return "senddown"
            if key in (353, 260):  # 353 shift tab, 260 left arrow
                # go to edit the file path
                self.set_edit_focus()
                self.window.noutrefresh()
                curses.doupdate()
                return "edit"
            if key == 10:
                # submit the file
                self.send_btn.focus = False
                self.send_btn.draw()
                try:
                    filepath = pathlib.Path(self._newvalue.strip()).expanduser().resolve()
                    blobformat = ''.join(filepath.suffixes)
                    members = {self.name : (filepath, 0, blobformat)}
                    await self.vector.send_newBLOBVector(members=members)
                except Exception:
                    logger.exception("Exception report from BLOB send action")
                    self.window.addstr( self.startline+2, 1, "!! Invalid !!    ", curses.color_pair(3) )
                else:
                    self.window.addstr( self.startline+2, 1, " - Sending -     ", curses.color_pair(1) )
                    self.vector.state = 'Busy'
                    draw_timestamp_state(self.control, self.tstatewin, self.vector)
                    self.tstatewin.noutrefresh()
                self.window.noutrefresh()
                curses.doupdate()
                time.sleep(0.4)      # blocking, to avoid screen being changed while this time elapses
                self.window.addstr( self.startline+2, 1, "Filepath to send:" )
                self.window.noutrefresh()
                curses.doupdate()
                return "senddown"
        else:
            # name and send button not in focus, so these are due to arrows in edit_text
            if key in (339, 259):  # 339 page up, 259 up arrow
                # go previous member widget edit field
                return "editup"
            if key in (338, 258):  # 338 page down, 258 down arrow
                # go to next widget edit field
                return "editdown"


    async def inputfield(self):
        "Input text, set it into self._newvalue"
        self.edit_txt.new_texteditor()

        while not self.control.stop:
            key = await self.keyinput()
            if key in ("Resize", "Messages", "Devices", "Vectors", "Stop"):
                return key
            if isinstance(key, tuple):
                if key in self.edit_txt:
                    continue
                return key
            if key in (9, 10):      # tab, enter
                self.send_btn.focus = True
                self.send_btn.draw()
                self.edit_txt.text = self._newvalue
                self.edit_txt.focus = False
                self.edit_txt.draw()
                self.window.noutrefresh()
                curses.doupdate()
                # goes back to the VectorScreen inputs method
                # which gets a key, and as this widget is still in focus
                # calls this widgets setkey method
                return
            if key == 353:           # shift tab
                self.name_btn.focus = True
                self.name_btn.draw()
                self.edit_txt.text = self._newvalue
                self.edit_txt.focus = False
                self.edit_txt.draw()
                self.window.noutrefresh()
                curses.doupdate()
                # goes back to the VectorScreen inputs method
                # which gets a key, and as this widget is still in focus
                # calls this widgets setkey method
                return
            if key in (339, 338, 259, 258):    # 339 page up, 338 page down, 259 up arrow, 258 down arrow
                self.edit_txt.text = self._newvalue
                return key
            self._newvalue = self.edit_txt.gettext(key)
            self.edit_txt.draw()
            self.window.noutrefresh()
            self.edit_txt.movecurs()
            curses.doupdate()




import collections, sys, time, json

from datetime import datetime, timezone

import asyncio

import xml.etree.ElementTree as ET

from .propertymembers import SwitchMember, LightMember, TextMember, NumberMember, BLOBMember, ParseException


class Vector(collections.UserDict):

    """This class is the parent of the PropertyVector class, which in turn
       is the parent of SwitchVector, LightVector, TextVector, NumberVector
       and BLOBVector classes.
       It is a mapping of membername to member value.
       """

    def __init__(self, name, label, group, state, timestamp, message):
        super().__init__()

        # UserDict will create self.data which will be a dictionary of
        # member name to member this vector owns

        self.name = name
        self.label = label
        self.group = group
        self._state = state
        self.timestamp = timestamp
        self.message = message
        self.vectortype = self.__class__.__name__
        self.devicename = None
        self._rule = None
        self._perm = None
        self.timeout = 0.0
        # if self.enable is False, this property is 'deleted'
        self.enable = True


    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        if not value in ('Idle','Ok','Busy','Alert'):
            raise ValueError("Invalid state given")
        self._state = value

    @property
    def rule(self):
        return self._rule

    @rule.setter
    def rule(self, value):
        if value not in ('OneOfMany', 'AtMostOne', 'AnyOfMany'):
            raise ValueError("Invalid rule given")
        self._rule = value

    @property
    def perm(self):
        return self._perm

    @perm.setter
    def perm(self, value):
        if value not in ('ro', 'wo', 'rw'):
            raise ValueError("Invalid permission given")
        self._perm = value

    def __setitem__(self, membername, value):
        self.data[membername].membervalue = value

    def __getitem__(self, membername):
        return self.data[membername].membervalue

    def members(self):
        "Returns a dictionary of member objects"
        return self.data

    def memberlabel(self, membername):
        "Returns the member label, given a member name"
        return self.data[membername].label


class SnapVector(Vector):
    """This object is used as a snapshot of this vector.
       Should you use the ipyclient.snapshot method to create a snapshot,
       the snapshot device will contain objects of this vector class. Which
       does not have the methods of SwitchVector, etc..
       This allows the snapshot to be read without risk of creating any
       side effects."""


    def __init__(self, name, label, group, state, timestamp, message,
                       vectortype, devicename, enable, data):
        super().__init__(name, label, group, state, timestamp, message)
        self.vectortype = vectortype
        self.devicename = devicename
        self.enable = enable
        for membername, member in data.items():
            self.data[membername] = member._snapshot()

    def dictdump(self):
        "Returns a dictionary of this vector"
        vecdict = {}
        memdict = {}
        for membername, member in self.data.items():
            memdict[membername] = member.dictdump()
        vecdict = {"vectortype":self.vectortype,
                   "name":self.name,
                   "devicename":self.devicename,
                   "label":self.label,
                   "message":self.message,
                   "enable":self.enable,
                   "group":self.group,
                   "state":self.state,
                   "timeout":self.timeout,
                   "timestamp":self.timestamp.isoformat(sep='T')}
        if self.rule:
            vecdict["rule"] = self.rule
        if self.perm:
            vecdict["perm"] = self.perm

        vecdict["members"] = memdict
        return vecdict

    def dumps(self, indent=None, separators=None):
        "Returns a JSON string of the snapshot."
        return json.dumps(self.dictdump(), indent=indent, separators=separators)


    def dump(self, fp, indent=None, separators=None):
        """Serialize the snapshot as a JSON formatted stream to fp, a file-like object.
           This uses the Python json module which always produces str objects, not bytes
           objects. Therefore, fp.write() must support str input."""
        return json.dump(self.dictdump(), fp, indent=indent, separators=separators)




class PropertyVector(Vector):
    "Inherits from Vector, and is the parent class of SwitchVector etc.."

    def __init__(self, name, label, group, state, timestamp, message, device, client):
        super().__init__(name, label, group, state, timestamp, message)
        self._client = client
        self.device = device
        self.devicename = device.devicename

        self._timer = False   # Set true when a timer is going after a newvector is sent
                              # set False when a setvector is received
        self._newtimer = 0    # Set to time.time() when a new vector is sent


    def checktimedout(self, nowtime):
        "Returns True if a timedout has occured, False otherwise"

        if not self._client.timeout_enable:
            self._timer = False
        if not self._timer:
            return False
        # so timer is running
        if self.timeout > self._client.vector_timeout_max:
            t = self._client.vector_timeout_max
        elif self.timeout < self._client.vector_timeout_min:
            t = self._client.vector_timeout_min
        else:
            t = self.timeout
        if nowtime > self._newtimer + t:
            # timed out
            self._timer = False
            return True
        return False

    def checkvalue(self, value, allowed):
        "allowed is a list of values, checks if value is in it"
        if value not in allowed:
            raise ParseException(f"Invalid value:{value}")
        return value

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._state = self.checkvalue(value, ['Idle','Ok','Busy','Alert'])


    def __setitem__(self, membername, value):
        "Members are added by being learnt from the driver, they cannot be manually added"
        raise KeyError

    def __getitem__(self, membername):
        return self.data[membername].membervalue

    def _setvector(self, event):
        "Updates this vector with new values after a set... vector has been received"
        if not self.enable:
            # this property does not exist
            return
        if event.state:
            self.state = event.state
        if event.timestamp:
            self.timestamp = event.timestamp
        if event.message:
            self.message = event.message
        if hasattr(event, 'timeout'):
            if not self.timeout is None:
                self.timeout = event.timeout
        for membername, membervalue in event.items():
            if membername in self.data:
                member = self.data[membername]
                member.membervalue = membervalue
        # turn off timer if all updates are successful
        self._timer = False


    def snapshot(self):
        """Take a snapshot of the vector and returns an object which is a restricted copy
           of the current state of the vector.
           Vector methods for sending data will not be available.
           This copy will not be updated by events. This is provided so that you can
           handle the vector data, without fear of the value changing."""

        snapvector = SnapVector(self.name, self.label, self.group, self.state, self.timestamp, self.message,
                                self.vectortype, self.devicename, self.enable, self.data)
        snapvector.timeout = self.timeout
        snapvector._rule = self._rule
        snapvector._perm = self._perm
        return snapvector



class SwitchVector(PropertyVector):

    """A SwitchVector sends and receives one or more members with values 'On' or 'Off'. It
       also has the extra attribute 'rule' which can be one of 'OneOfMany', 'AtMostOne', 'AnyOfMany'.
       These are hints to the client how to display the switches in the vector.

       OneOfMany - of the SwitchMembers in this vector, one (and only one) must be On.

       AtMostOne - of the SwitchMembers in this vector, one or none can be On.

       AnyOfMany - multiple switch members can be On.
       """

    def __init__(self, event):
        super().__init__(event.vectorname, event.label, event.group, event.state,
                         event.timestamp, event.message, event.device, event._client)
        self._perm = event.perm
        self._rule = event.rule
        self.timeout = event.timeout
        # self.data is a dictionary of switch name : switchmember
        # create  members
        for membername, membervalue in event.items():
            self.data[membername] = SwitchMember(membername, event.memberlabels[membername], membervalue)

    @property
    def perm(self):
        return self._perm

    @perm.setter
    def perm(self, value):
        self._perm = self.checkvalue(value, ['ro','wo','rw'])

    @property
    def rule(self):
        return self._rule

    @rule.setter
    def rule(self, value):
        self._rule = self.checkvalue(value, ['OneOfMany','AtMostOne','AnyOfMany'])


    def _defvector(self, event):
        "Updates this vector with new values after a def... vector has been received"
        self._timer = False
        if event.label:
            self.label = event.label
        if event.group:
            self.group = event.group
        if event.perm:
            self.perm = event.perm
        if event.rule:
            self.rule = event.rule
        if event.state:
            self.state = event.state
        if event.timestamp:
            self.timestamp = event.timestamp
        if event.message:
            self.message = event.message
        self.timeout = event.timeout
        # create  members
        for membername, membervalue in event.items():
            if membername in self.data:
                # update existing member
                if event.memberlabels[membername]:
                    self.data[membername].label = event.memberlabels[membername]
                self.data[membername].membervalue = membervalue
            else:
                # create new member
                self.data[membername] = SwitchMember(membername, event.memberlabels[membername], membervalue)
        self.enable = True

    def _newSwitchVector(self, timestamp=None, members={}):
        "Creates the xmldata for sending a newSwitchVector"
        if not self.enable:
            return
        if timestamp is None:
            timestamp = datetime.now(tz=timezone.utc)
        if not isinstance(timestamp, datetime):
            # invalid timestamp given
            return
        if not (timestamp.tzinfo is None):
            if timestamp.tzinfo == timezone.utc:
                timestamp = timestamp.replace(tzinfo = None)
            else:
                # invalid timestamp
                return
        # timestamp has no tzinfo so isoformat does not include timezone info
        self.state = 'Busy'
        xmldata = ET.Element('newSwitchVector')
        xmldata.set("device", self.devicename)
        xmldata.set("name", self.name)
        xmldata.set("timestamp", timestamp.isoformat(sep='T'))
        # set member values to send
        sendvalues = {}
        for membername, value in members.items():
            # check this membername exists
            if membername in self:
                sendvalues[membername] = value
        # for rule 'OneOfMany' the standard indicates 'Off' should precede 'On'
        # so make all 'On' values last
        Offswitches = []
        Onswitches = []
        for mname, value in sendvalues.items():
            if value == 'Off':
                # create list of (memberswitch, new value) tuples
                Offswitches.append( ( self.data[mname], value ) )
            elif value == 'On':
               Onswitches.append( ( self.data[mname], value ) )
        for switch,value in Offswitches:
            xmldata.append(switch.oneswitch(value))
        for switch, value in Onswitches:
            xmldata.append(switch.oneswitch(value))
        return xmldata


    async def send_newSwitchVector(self, timestamp=None, members={}):
        """Transmits the vector (newSwitchVector) and the members given in the members
           dictionary which consists of member names:values to be sent.
           The values should be strings of either On or Off.
           This method will encode and transmit the xml, and change the vector state to busy.
           If no timestamp is given, a current UTC time will be created."""
        xmldata = self._newSwitchVector(timestamp, members)
        if xmldata is None:
            return
        self._timer = True
        self._newtimer = time.time()
        await self._client.send(xmldata)


class LightVector(PropertyVector):

    """A LightVector is an instrument indicator, and has one or more members
       with values 'Idle', 'Ok', 'Busy' or 'Alert'. In general a client will
       indicate this state with different colours.

       This class has no 'send_newLightVector method, since lights are read-only"""

    def __init__(self, event):
        super().__init__(event.vectorname, event.label, event.group, event.state,
                         event.timestamp, event.message, event.device, event._client)
        self._perm = "ro"
        # self.data is a dictionary of light name : lightmember
        # create  members
        for membername, membervalue in event.items():
            self.data[membername] = LightMember(membername, event.memberlabels[membername], membervalue)

    @property
    def perm(self):
        return "ro"

    @perm.setter
    def perm(self, value):
        pass

    def checktimedout(self, nowtime):
        "As ro, always returns False"
        return False

    def _defvector(self, event):
        "Updates this vector with new values after a def... vector has been received"
        if event.label:
            self.label = event.label
        if event.group:
            self.group = event.group
        if event.state:
            self.state = event.state
        if event.timestamp:
            self.timestamp = event.timestamp
        if event.message:
            self.message = event.message
        # create  members
        for membername, membervalue in event.items():
            if membername in self.data:
                # update existing member
                if event.memberlabels[membername]:
                    self.data[membername].label = event.memberlabels[membername]
                self.data[membername].membervalue = membervalue
            else:
                # create new member
                self.data[membername] = LightMember(membername, event.memberlabels[membername], membervalue)
        self.enable = True

    def snapshot(self):
        """Take a snapshot of the vector and returns an object which is a restricted copy
           of the current state of the vector.
           Vector methods for sending data will not be available.
           This copy will not be updated by events. This is provided so that you can
           handle the vector data, without fear of the value changing."""

        snapvector = SnapVector(self.name, self.label, self.group, self.state, self.timestamp, self.message,
                                self.vectortype, self.devicename, self.enable, self.data)
        snapvector._perm = "ro"
        return snapvector



class TextVector(PropertyVector):

    """A TextVector is used to send and receive text between instrument and client."""


    def __init__(self, event):
        super().__init__(event.vectorname, event.label, event.group, event.state,
                         event.timestamp, event.message, event.device, event._client)
        self._perm = event.perm
        self.timeout = event.timeout
        # self.data is a dictionary of text name : textmember
        # create  members
        for membername, membervalue in event.items():
            self.data[membername] = TextMember(membername, event.memberlabels[membername], membervalue)

    @property
    def perm(self):
        return self._perm

    @perm.setter
    def perm(self, value):
        self._perm = self.checkvalue(value, ['ro','wo','rw'])

    def _defvector(self, event):
        "Updates this vector with new values after a def... vector has been received"
        self._timer = False
        if event.label:
            self.label = event.label
        if event.group:
            self.group = event.group
        if event.perm:
            self.perm = event.perm
        if event.state:
            self.state = event.state
        if event.timestamp:
            self.timestamp = event.timestamp
        if event.message:
            self.message = event.message
        self.timeout = event.timeout
        # create  members
        for membername, membervalue in event.items():
            if membername in self.data:
                # update existing member
                if event.memberlabels[membername]:
                    self.data[membername].label = event.memberlabels[membername]
                self.data[membername].membervalue = membervalue
            else:
                # create new member
                self.data[membername] = TextMember(membername, event.memberlabels[membername], membervalue)
        self.enable = True


    def _newTextVector(self, timestamp=None, members={}):
        "Creates the xmldata for sending a newTextVector"
        if not self.enable:
            return
        if timestamp is None:
            timestamp = datetime.now(tz=timezone.utc)
        if not isinstance(timestamp, datetime):
            # invalid timestamp given
            return
        if not (timestamp.tzinfo is None):
            if timestamp.tzinfo == timezone.utc:
                timestamp = timestamp.replace(tzinfo = None)
            else:
                # invalid timestamp
                return
        # timestamp has no tzinfo so isoformat does not include timezone info
        self.state = 'Busy'
        xmldata = ET.Element('newTextVector')
        xmldata.set("device", self.devicename)
        xmldata.set("name", self.name)
        xmldata.set("timestamp", timestamp.isoformat(sep='T'))
        # set member values to send
        for membername, textmember in self.data.items():
            if membername in members:
                xmldata.append(textmember.onetext(members[membername]))
            else:
                xmldata.append(textmember.onetext(textmember.membervalue))
        return xmldata


    async def send_newTextVector(self, timestamp=None, members={}):
        """Transmits the vector (newTextVector) with members and values.
           members is a dictionary of membernames:text string values.
           The spec requires text vectors to be sent with all members, so if the given
           members dictionary only includes changed values, the remaining members with
           unchanged values will still be sent.
           This method will transmit the vector and change the vector state to busy.
           If no timestamp is given, a current UTC time will be created."""
        xmldata = self._newTextVector(timestamp, members)
        if xmldata is None:
            return
        self._timer = True
        self._newtimer = time.time()
        await self._client.send(xmldata)



class SnapNumberVector(SnapVector):

    def getfloatvalue(self, membername):
        "Given a membername of this vector, returns the number as a float"
        if membername not in self:
            raise KeyError(f"Unrecognised member: {membername}")
        member = self.data[membername]
        return member.getfloatvalue()

    def getformattedvalue(self, membername):
        "Given a membername of this vector, returns the number as a formatted string"
        if membername not in self:
            raise KeyError(f"Unrecognised member: {membername}")
        member = self.data[membername]
        return member.getformattedvalue()




class NumberVector(PropertyVector):
    """A NumberVector is used to send and receive numbers between instrument and client.
       As data is received, this vector is a mapping of membername:membervalue where membervalue
       is the string of the number taken from the received xml.
       The INDI spec defines a number of formats, including degrees:minutes:seconds so
       this class includes methods to obtain the number as a float, and to create a string
       formatted as the particular member requires.
       The member objects contain further information label, format spec, minimum, maximum and step size.
       To obtain the member object, as opposed to the member value, use the members() method."""

    def __init__(self, event):
        super().__init__(event.vectorname, event.label, event.group, event.state,
                         event.timestamp, event.message, event.device, event._client)
        self._perm = event.perm
        self.timeout = event.timeout
        # self.data is a dictionary of number name : numbermember
        # create  members
        for membername, membervalue in event.items():
            self.data[membername] = NumberMember(membername, *event.memberlabels[membername], membervalue)

    def getfloatvalue(self, membername):
        "Given a membername of this vector, returns the number as a float"
        if membername not in self:
            raise KeyError(f"Unrecognised member: {membername}")
        member = self.data[membername]
        return member.getfloatvalue()

    def getformattedvalue(self, membername):
        "Given a membername of this vector, returns the number as a formatted string"
        if membername not in self:
            raise KeyError(f"Unrecognised member: {membername}")
        member = self.data[membername]
        return member.getformattedvalue()

    @property
    def perm(self):
        return self._perm

    @perm.setter
    def perm(self, value):
        self._perm = self.checkvalue(value, ['ro','wo','rw'])

    def _defvector(self, event):
        "Updates this vector with new values after a def... vector has been received"
        self._timer = False
        if event.label:
            self.label = event.label
        if event.group:
            self.group = event.group
        if event.perm:
            self.perm = event.perm
        if event.state:
            self.state = event.state
        if event.timestamp:
            self.timestamp = event.timestamp
        if event.message:
            self.message = event.message
        self.timeout = event.timeout
        # create  members
        for membername, membervalue in event.items():
            if membername in self.data:
                # update existing member
                if event.memberlabels[membername]:
                    member = self.data[membername]
                    member.label = event.memberlabels[membername][0]
                    member.format = event.memberlabels[membername][1]
                    member.min = event.memberlabels[membername][2]
                    member.max = event.memberlabels[membername][3]
                    member.step = event.memberlabels[membername][4]
                self.data[membername].membervalue = membervalue
            else:
                # create new member
                self.data[membername] = NumberMember(membername, *event.memberlabels[membername], membervalue)
        self.enable = True


    def _newNumberVector(self, timestamp=None, members={}):
        "Creates the xmldata for sending a newNumberVector"
        if not self.enable:
            return
        if timestamp is None:
            timestamp = datetime.now(tz=timezone.utc)
        if not isinstance(timestamp, datetime):
            # invalid timestamp given
            return
        if not (timestamp.tzinfo is None):
            if timestamp.tzinfo == timezone.utc:
                timestamp = timestamp.replace(tzinfo = None)
            else:
                # invalid timestamp
                return
        # timestamp has no tzinfo so isoformat does not include timezone info
        self.state = 'Busy'
        xmldata = ET.Element('newNumberVector')
        xmldata.set("device", self.devicename)
        xmldata.set("name", self.name)
        xmldata.set("timestamp", timestamp.isoformat(sep='T'))
        # set member values to send
        for membername, numbermember in self.data.items():
            if membername in members:
                value = members[membername]
                if isinstance(value, float) or isinstance(value, int):
                    value = str(value)
                xmldata.append(numbermember.onenumber(value))
            else:
                xmldata.append(numbermember.onenumber(numbermember.membervalue))
        return xmldata

    async def send_newNumberVector(self, timestamp=None, members={}):
        """Transmits the vector (newNumberVector) with members and values.
           members is a dictionary of membernames:number values, the values can be
           integers, floats or strings, if not strings they will be converted to strings.
           The spec requires number vectors to be sent with all members, so if the given
           members dictionary only includes changed values, the remaining members with
           unchanged values will still be sent.
           This method will transmit the vector and change the vector state to busy.
           If no timestamp is given, a current UTC time will be created."""
        xmldata = self._newNumberVector(timestamp, members)
        if xmldata is None:
            return
        self._timer = True
        self._newtimer = time.time()
        await self._client.send(xmldata)

    def snapshot(self):
        """Take a snapshot of the vector and returns an object which is a restricted copy
           of the current state of the vector.
           Vector methods for sending data will not be available.
           This copy will not be updated by events. This is provided so that you can
           handle the vector data, without fear of the value changing."""

        snapvector = SnapNumberVector(self.name, self.label, self.group, self.state, self.timestamp, self.message,
                                self.vectortype, self.devicename, self.enable, self.data)
        snapvector.timeout = self.timeout
        snapvector._perm = self._perm
        return snapvector





class BLOBVector(PropertyVector):

    """A BLOBVector is used to send and receive Binary Large Objects between instrument and client.
       As data is received this vector will be a mapping of membername to membervalue where membervalue
       will be a binary string of the received BLOB.  This binary string has been decoded from the
       received XML and from the b64 encoding used.
       The member object contains further information, label, blobsize and blobformat.
       To obtain the member object, as opposed to the member value, use the members() method."""

    def __init__(self, event):
        super().__init__(event.vectorname, event.label, event.group, event.state,
                         event.timestamp, event.message, event.device, event._client)
        self._perm = event.perm
        self.timeout = event.timeout
        # self.data is a dictionary of blob name : blobmember
        # create  members
        for membername, label in event.memberlabels.items():
            self.data[membername] = BLOBMember(membername, label)

    def set_blobsize(self, membername, blobsize):
        "Used when an event is received to set the member blobsize"
        if not isinstance(blobsize, int):
            return
        if membername in self.data:
            member = self.data[membername]
        else:
            return
        member.blobsize = blobsize

    def set_blobformat(self, membername, blobformat):
        """Sets the blobformat attribute in the blob member."""
        if membername in self.data:
            member = self.data[membername]
        else:
            return
        member.blobformat = blobformat

    @property
    def perm(self):
        return self._perm

    @perm.setter
    def perm(self, value):
        self._perm = self.checkvalue(value, ['ro','wo','rw'])

    def _defvector(self, event):
        "Updates this vector with new values after a def... vector has been received"
        self._timer = False
        if event.label:
            self.label = event.label
        if event.group:
            self.group = event.group
        if event.perm:
            self.perm = event.perm
        if event.state:
            self.state = event.state
        if event.timestamp:
            self.timestamp = event.timestamp
        if event.message:
            self.message = event.message
        self.timeout = event.timeout
        # create  members
        for membername, label in event.memberlabels.items():
            if membername in self.data:
                # update existing member
                self.data[membername].label = label
            else:
                # create new member
                self.data[membername] = BLOBMember(membername, label)
        self.enable = True


    def _setvector(self, event):
        "Updates this vector with new values after a setBLOBvector has been received"
        self._timer = False
        if not self.enable:
            # this property does not exist
            return
        super()._setvector(event)
        # set each members size and format
        # using event.sizeformat[membername] = (membersize, memberformat)
        for membername in event.keys():
            membersize, memberformat = event.sizeformat[membername]
            self.set_blobsize(membername, membersize)
            self.set_blobformat(membername, memberformat)


    def _newBLOBVector(self, timestamp=None, members={}):
        "Creates the xmldata for sending a newBLOBVector"
        if not self.enable:
            return
        if timestamp is None:
            timestamp = datetime.now(tz=timezone.utc)
        if not isinstance(timestamp, datetime):
            # invalid timestamp given
            return
        if not (timestamp.tzinfo is None):
            if timestamp.tzinfo == timezone.utc:
                timestamp = timestamp.replace(tzinfo = None)
            else:
                # invalid timestamp
                return
        # timestamp has no tzinfo so isoformat does not include timezone info
        self.state = 'Busy'
        xmldata = ET.Element('newBLOBVector')
        xmldata.set("device", self.devicename)
        xmldata.set("name", self.name)
        xmldata.set("timestamp", timestamp.isoformat(sep='T'))
        # set member values to send
        for membername, blobmember in self.data.items():
            if membername in members:
                # expand (value, blobsize, blobformat) to required arguments
                xmldata.append(blobmember.oneblob(*members[membername]))
        return xmldata

    async def send_newBLOBVector(self, timestamp=None, members={}):
        """Transmits the vector (newBLOBVector) with new BLOB members
           This method will transmit the vector and change the vector state to busy.
           The members dictionary should be {membername:(value, blobsize, blobformat)}
           The value could be a bytes object, a pathlib.Path, a string path to a file or a file-like object.
           If blobsize of zero is used, the size value sent will be set to the number of bytes
           in the BLOB. The INDI standard specifies the size should be that of the BLOB
           before any compression, therefore if you are sending a compressed file, you
           should set the blobsize prior to compression.
           blobformat should be a file extension, such as '.png'"""
        xmldata = self._newBLOBVector(timestamp, members)
        if xmldata is None:
            return
        self._timer = True
        self._newtimer = time.time()
        await self._client.send(xmldata)




import xml.etree.ElementTree as ET

import sys, pathlib

from base64 import standard_b64encode


class ParseException(Exception):
    "Raised if an error occurs when parsing received data"
    pass


def getfloat(value):
    """The INDI spec specifies several different number formats, given a number
       in any of these formats, this returns a float.
       If an error occurs while parsing the number, a TypeError exception is raised."""
    try:
        if isinstance(value, float):
            return value
        if isinstance(value, int):
            return float(value)
        if not isinstance(value, str):
            raise TypeError
        # negative is True, if the value is negative
        value = value.strip()
        negative = value.startswith("-")
        if negative:
            value = value.lstrip("-")
        # Is the number provided in sexagesimal form?
        if value == "":
            parts = ["0", "0", "0"]
        elif " " in value:
            parts = value.split(" ")
        elif ":" in value:
            parts = value.split(":")
        elif ";" in value:
            parts = value.split(";")
        else:
            # not sexagesimal
            parts = [value, "0", "0"]
        if len(parts) > 3:
            raise TypeError
        # Any missing parts should have zero
        if len(parts) == 1:
            parts.append("0")
            parts.append("0")
        if len(parts) == 2:
            parts.append("0")
        assert len(parts) == 3
        # a part could be empty string, ie if 2:5: is given
        numbers = list(float(x) if x else 0.0 for x in parts)
        floatvalue = numbers[0] + (numbers[1]/60) + (numbers[2]/3600)
        if negative:
            floatvalue = -1 * floatvalue
    except Exception:
        raise TypeError("Unable to parse number value")
    return floatvalue


class Member():
    """This class is the parent of further member classes."""

    def __init__(self, name, label=None, membervalue=None):
        self.name = name
        if label:
            self.label = label
        else:
            self.label = name
        self._membervalue = membervalue

    @property
    def membervalue(self):
        return self._membervalue

    @membervalue.setter
    def membervalue(self, value):
        self._membervalue = value


class SnapMember(Member):

    """Should you use the ipyclient.snapshot method to create a snapshot,
       the snapshot members for Switch, Light and Text will be objects
       of this class."""

    def dictdump(self):
        "Returns a dictionary of this member"
        return {"label": self.label,
                "value": self._membervalue}


class SwitchMember(Member):
    """A SwitchMember can only have one of 'On' or 'Off' values"""

    def __init__(self, name, label=None, membervalue="Off"):
        super().__init__(name, label, membervalue)
        if membervalue not in ('On', 'Off'):
            raise ParseException(f"Invalid value {membervalue}, should be On or Off")

    @property
    def membervalue(self):
        return self._membervalue

    @membervalue.setter
    def membervalue(self, value):
        if not value:
            raise ParseException("No value given, should be On or Off")
        newvalue = self.checkvalue(value, ['On', 'Off'])
        if self._membervalue != newvalue:
            self._membervalue = newvalue

    def checkvalue(self, value, allowed):
        "allowed is a list of values, checks if value is in it"
        if value not in allowed:
            raise ParseException(f"Invalid value:{value}")
        return value

    def _snapshot(self):
        snapmember = SnapMember(self.name, self.label, self._membervalue)
        return snapmember

    def oneswitch(self, newvalue):
        """Returns xml of a oneSwitch with the new value to send"""
        xmldata = ET.Element('oneSwitch')
        xmldata.set("name", self.name)
        xmldata.text = newvalue
        return xmldata


class LightMember(Member):
    """A LightMember can only have one of 'Idle', 'Ok', 'Busy' or 'Alert' values"""

    def __init__(self, name, label=None, membervalue="Idle"):
        super().__init__(name, label, membervalue)
        if membervalue not in ('Idle','Ok','Busy','Alert'):
            raise ParseException(f"Invalid light value {membervalue}")

    @property
    def membervalue(self):
        return self._membervalue

    @membervalue.setter
    def membervalue(self, value):
        if not value:
            raise ParseException("No light value given")
        newvalue = self.checkvalue(value, ['Idle','Ok','Busy','Alert'])
        if self._membervalue != newvalue:
            self._membervalue = newvalue

    def checkvalue(self, value, allowed):
        "allowed is a list of values, checks if value is in it"
        if value not in allowed:
            raise ParseException(f"Invalid value:{value}")
        return value

    def _snapshot(self):
        snapmember = SnapMember(self.name, self.label, self._membervalue)
        return snapmember


class TextMember(Member):
    """Contains a text string"""

    def __init__(self, name, label=None, membervalue=""):
        super().__init__(name, label, membervalue)
        if not isinstance(membervalue, str):
            raise ParseException("The text value should be a string")

    @property
    def membervalue(self):
        return self._membervalue

    @membervalue.setter
    def membervalue(self, value):
        if not isinstance(value, str):
            raise ParseException("The text value should be a string")
        if self._membervalue != value:
            self._membervalue = value

    def _snapshot(self):
        snapmember = SnapMember(self.name, self.label, self._membervalue)
        return snapmember

    def onetext(self, newvalue):
        """Returns xml of a oneText"""
        xmldata = ET.Element('oneText')
        xmldata.set("name", self.name)
        xmldata.text = newvalue
        return xmldata


class ParentNumberMember(Member):

    """This class inherits from Member and is the parent of the NumberMember class.
    """


    def __init__(self, name, label=None, format='', min='0', max='0', step='0', membervalue='0'):
        super().__init__(name, label, membervalue)
        self.format = format
        self.min = min
        self.max = max
        self.step = step

    def getfloat(self, value):
        """The INDI spec specifies several different number formats, this method returns
           the given value as a float.
           If an error occurs while parsing the number, a TypeError exception is raised."""
        return getfloat(value)


    def getfloatvalue(self):
        """The INDI spec allows a number of different number formats, this method returns
           this members value as a float.
           If an error occurs while parsing the number, a TypeError exception is raised."""
        return getfloat(self._membervalue)


    def getformattedvalue(self):
        """This method returns this members value as a formatted string."""
        return self.getformattedstring(self._membervalue)


    def getformattedstring(self, value):
        """Given a number this returns a formatted string"""
        try:
            value = getfloat(value)
            if (not self.format.startswith("%")) or (not self.format.endswith("m")):
                return self.format % value
            # sexagesimal
            # format string is of the form  %<w>.<f>m
            w,f = self.format.split(".")
            w = w.lstrip("%")
            f = f.rstrip("m")

            if value<0:
                negative = True
            else:
                negative = False
            absvalue = abs(value)

            # get integer part and fraction part
            degrees = int(absvalue)
            minutes = (absvalue - degrees) * 60.0

            if f == "3":   # three fractional values including the colon ":mm"
                # create nearest integer minutes
                minutes = round(minutes)
                if minutes == 60:
                    minutes = 0
                    degrees = degrees + 1
                valstring = f"{'-' if negative else ''}{degrees}:{minutes:02d}"
                # w is the overall length of the string, prepend with spaces to make the length up to w
                if w:
                    return valstring.rjust(int(w), ' ')
                # it is possible w is an empty string
                return valstring

            if f == "5":  # five fractional values including the colon and decimal point ":mm.m"
                minutes = round(minutes,1)
                if minutes == 60.0:
                    minutes = 0.0
                    degrees = degrees + 1
                valstring = f"{'-' if negative else ''}{degrees}:{minutes:04.1f}"
                if w:
                    return valstring.rjust(int(w), ' ')
                return valstring

            integerminutes = int(minutes)
            seconds = (minutes - integerminutes) * 60.0

            if f == "6":    # six fractional values including two colons ":mm:ss"
                seconds = round(seconds)
                if seconds == 60:
                    seconds = 0
                    integerminutes = integerminutes + 1
                    if integerminutes == 60:
                        integerminutes = 0
                        degrees = degrees + 1
                valstring = f"{'-' if negative else ''}{degrees}:{integerminutes:02d}:{seconds:02d}"
                if w:
                    return valstring.rjust(int(w), ' ')
                return valstring


            if f == "8":    # eight fractional values including two colons and decimal point ":mm:ss.s"
                seconds = round(seconds,1)
                if seconds == 60.0:
                    seconds = 0.0
                    integerminutes = integerminutes + 1
                    if integerminutes == 60:
                        integerminutes = 0
                        degrees = degrees + 1
                valstring = f"{'-' if negative else ''}{degrees}:{integerminutes:02d}:{seconds:04.1f}"
                if w:
                    return valstring.rjust(int(w), ' ')
                return valstring

            fn = int(f)
            if fn>8 and fn<15:    # make maximum of 14
                seconds = round(seconds,1)
                if seconds == 60.0:
                    seconds = 0.0
                    integerminutes = integerminutes + 1
                    if integerminutes == 60:
                        integerminutes = 0
                        degrees = degrees + 1
                valstring = f"{'-' if negative else ''}{degrees}:{integerminutes:02d}:{seconds:0{fn-4}.{fn-7}f}"
                if w:
                    return valstring.rjust(int(w), ' ')
                return valstring

        except Exception:
            raise TypeError("Unable to parse number value")

        # no other options accepted
        raise TypeError("Unable to process number format")


class SnapNumberMember(ParentNumberMember):

    """Should you use the ipyclient.snapshot method to create a snapshot,
       the snapshot members for Numbers will be objects of this class."""

    def dictdump(self):
        "Returns a dictionary of this member"
        return {"label": self.label,
                "format": self.format,
                "min": self.min,
                "max": self.max,
                "step": self.step,
                "value": self._membervalue,
                "floatvalue": self.getfloatvalue(),
                "formattedvalue": self.getformattedvalue()
                }


class NumberMember(ParentNumberMember):
    """Contains a number, the attributes inform the client how the number should be
       displayed.
    """

    def __init__(self, name, label=None, format='', min='0', max='0', step='0', membervalue='0'):
        super().__init__(name, label, format, min, max, step, membervalue)
        self.format = format
        if isinstance(min, str):
            self.min = min
        else:
            self.min = str(min)
        if isinstance(max, str):
            self.max = max
        else:
            self.max = str(max)
        if isinstance(step, str):
            self.step = step
        else:
            self.step = str(step)
        if isinstance(membervalue, str):
            value = membervalue
        else:
            value = str(membervalue)
        try:
            # test a float can be created from this value
            self._floatvalue = getfloat(value)
        except Exception:
            raise ParseException("Cannot parse number received.")
        self._membervalue = value

    @property
    def membervalue(self):
        return self._membervalue

    @membervalue.setter
    def membervalue(self, value):
        if not isinstance(value, str):
            value = str(value)
        if self._membervalue == value:
            return
        try:
            # test a float can be created from this membervalue
            # and save the float
            self._floatvalue = getfloat(value)
        except Exception:
            raise ParseException("Cannot parse number received")
        self._membervalue = value


    def getfloatvalue(self):
        """The INDI spec allows a number of different number formats, this method returns
           this members value as a float."""
        return self._floatvalue


    def onenumber(self, newvalue):
        """Returns xml of a oneNumber"""
        if isinstance(newvalue, str):
            value = newvalue
        else:
            value = str(newvalue)
        xmldata = ET.Element('oneNumber')
        xmldata.set("name", self.name)
        xmldata.text = value
        return xmldata

    def _snapshot(self):
        snapmember = SnapNumberMember(self.name, self.label, self.format, self.min, self.max, self.step, self._membervalue)
        return snapmember


class ParentBLOBMember(Member):

    """This class inherits from Member and is the parent of the BLOBMember class.
    """


    def __init__(self, name, label=None, blobsize=0, blobformat='', membervalue=None):
        super().__init__(name, label, membervalue)
        self.blobsize = blobsize
        self.blobformat = blobformat


class SnapBLOBMember(ParentBLOBMember):

    """Should you use the ipyclient.snapshot method to create a snapshot,
       the snapshot members for BLOBs will be objects of this class."""

    def dictdump(self):
        "Returns a dictionary of this member, value is always None"
        return {"label": self.label,
                "blobsize": self.blobsize,
                "blobformat": self.blobformat,
                "value": None}


class BLOBMember(ParentBLOBMember):
    """Contains a 'binary large object' such as an image."""

    def __init__(self, name, label=None, blobsize=0, blobformat='', membervalue=None):
        super().__init__(name, label, membervalue)
        if not isinstance(blobsize, int):
            raise ParseException("Blobsize must be given as an integer")
        # membervalue can be a byte string, path, string path or file like object
        self.blobsize = blobsize
        self.blobformat = blobformat

    @property
    def membervalue(self):
        return self._membervalue

    @membervalue.setter
    def membervalue(self, value):
        if not value:
            raise ParseException("No BLOB value given")
        self._membervalue = value


    def oneblob(self, newvalue, newsize, newformat):
        """Returns xml of a oneBLOB"""
        xmldata = ET.Element('oneBLOB')
        xmldata.set("name", self.name)
        xmldata.set("format", newformat)
        # the value set in the xmldata object should be a bytes object
        if isinstance(newvalue, bytes):
            bytescontent = newvalue
        elif isinstance(newvalue, pathlib.Path):
            try:
                bytescontent = newvalue.read_bytes()
            except Exception:
                raise ParseException("Unable to read the given file")
        elif hasattr(newvalue, "seek") and hasattr(newvalue, "read") and callable(newvalue.read):
            # a file-like object
            # set seek(0) so is read from start of file
            newvalue.seek(0)
            bytescontent = newvalue.read()
            newvalue.close()
            if not isinstance(bytescontent, bytes):
                raise ParseException("The read BLOB is not a bytes object")
            if bytescontent == b"":
                raise ParseException("The read BLOB value is empty")
        else:
            # could be a path to a file
            try:
                with open(newvalue, "rb") as fp:
                    bytescontent = fp.read()
            except Exception:
                raise ParseException("Unable to read the given file")
            if bytescontent == b"":
                raise ParseException("The read BLOB value is empty")
        if not newsize:
            newsize = len(bytescontent)
        xmldata.set("size", str(newsize))
        xmldata.text = standard_b64encode(bytescontent).decode("utf-8")
        return xmldata


    def _snapshot(self):
        snapmember = SnapBLOBMember(self.name, self.label, self.blobsize, self.blobformat, self._membervalue)
        return snapmember
